"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import unittest
from typing import (
    Generic,
)

from minos.common import (
    CUSTOM_TYPES,
    Decimal,
    Enum,
    Fixed,
    MissingSentinel,
    ModelRef,
)


class TestMissingSentinel(unittest.TestCase):
    def test_subclass(self):
        # noinspection PyTypeHints
        self.assertTrue(issubclass(MissingSentinel, Generic))


class TestFixed(unittest.TestCase):
    def test_subclass(self):
        # noinspection PyTypeHints
        self.assertTrue(issubclass(Enum, Generic))

    def test_symbols(self):
        fixed = Fixed(4)
        self.assertEqual(4, fixed.size)

    def test_default(self):
        fixed = Fixed(4)
        self.assertEqual(MissingSentinel, fixed.default)

    def test_namespace(self):
        fixed = Fixed(4)
        self.assertIsNone(fixed.namespace)

    def test_aliases(self):
        fixed = Fixed(4)
        self.assertIsNone(fixed.aliases)

    def test_repr(self):
        fixed = Fixed(4)
        self.assertEqual("Fixed(size=4)", repr(fixed))


class TestEnum(unittest.TestCase):
    def test_subclass(self):
        # noinspection PyTypeHints
        self.assertTrue(issubclass(Enum, Generic))

    def test_symbols(self):
        enum = Enum(["one", "two", "three"])
        self.assertEqual(["one", "two", "three"], enum.symbols)

    def test_default(self):
        enum = Enum(["one", "two", "three"])
        self.assertEqual(MissingSentinel, enum.default)

    def test_namespace(self):
        enum = Enum(["one", "two", "three"])
        self.assertIsNone(enum.namespace)

    def test_aliases(self):
        enum = Enum(["one", "two", "three"])
        self.assertIsNone(enum.aliases)

    def test_docs(self):
        enum = Enum(["one", "two", "three"])
        self.assertIsNone(enum.docs)

    def test_repr(self):
        enum = Enum(["one", "two", "three"])
        self.assertEqual("Enum(symbols=['one', 'two', 'three'])", repr(enum))


class TestDecimal(unittest.TestCase):
    def test_subclass(self):
        # noinspection PyTypeHints
        self.assertTrue(issubclass(Decimal, Generic))

    def test_precision(self):
        decimal = Decimal(6)
        self.assertEqual(6, decimal.precision)

    def test_scale(self):
        decimal = Decimal(6)
        self.assertEqual(0, decimal.scale)

    def test_default(self):
        decimal = Decimal(6)
        self.assertEqual(MissingSentinel, decimal.default)

    def test_aliases(self):
        decimal = Decimal(6)
        self.assertIsNone(decimal.aliases)

    def test_repr(self):
        decimal = Decimal(6)
        self.assertEqual("Decimal(precision=6, scale=0)", repr(decimal))


class TestModelRef(unittest.TestCase):
    def test_subclass(self):
        # noinspection PyTypeHints
        self.assertTrue(issubclass(ModelRef, Generic))

    def test_repr(self):
        ref = ModelRef()
        self.assertEqual("ModelRef()", repr(ref))


class TestTypesModule(unittest.TestCase):
    def test_custom_types(self):
        self.assertEqual(("Fixed", "Enum", "Decimal", "ModelRef",), CUSTOM_TYPES)


if __name__ == "__main__":
    unittest.main()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
import unittest

from minos.common import (
    MinosAvroProtocol,
    MinosAvroValuesDatabase,
    MinosProtocolException,
)


class TestMinosAvroProtocol(unittest.TestCase):
    def test_encoder_decoder_with_body_dict(self):
        headers = {"id": 123, "action": "get"}
        body = {"message": "this is a message"}
        data_return_bytes = MinosAvroProtocol.encode(headers, body)
        self.assertEqual(True, isinstance(data_return_bytes, bytes))

        # decode
        returned_dict = MinosAvroProtocol.decode(data_return_bytes)
        self.assertEqual("get", returned_dict["headers"]["action"])
        self.assertEqual(123, returned_dict["headers"]["id"])
        self.assertEqual("this is a message", returned_dict["body"]["message"])

    def test_encoder_decoder_with_body_string(self):
        headers = {"id": 123, "action": "get"}
        body = "this is a message"
        data_return_bytes = MinosAvroProtocol.encode(headers, body)
        self.assertEqual(True, isinstance(data_return_bytes, bytes))

        # decode
        returned_dict = MinosAvroProtocol.decode(data_return_bytes)
        self.assertEqual("get", returned_dict["headers"]["action"])
        self.assertEqual(123, returned_dict["headers"]["id"])
        self.assertEqual("this is a message", returned_dict["body"])

    def test_encoder_decoder_with_body_int(self):
        headers = {"id": 123, "action": "get"}
        body = 222
        data_return_bytes = MinosAvroProtocol.encode(headers, body)
        self.assertEqual(True, isinstance(data_return_bytes, bytes))

        # decode
        returned_dict = MinosAvroProtocol.decode(data_return_bytes)
        self.assertEqual("get", returned_dict["headers"]["action"])
        self.assertEqual(123, returned_dict["headers"]["id"])
        self.assertEqual(222, returned_dict["body"])

    def test_encoder_decoder_with_body_array(self):
        headers = {"id": 123, "action": "get"}
        body = ["testing", "testing2"]
        data_return_bytes = MinosAvroProtocol.encode(headers, body)
        self.assertEqual(True, isinstance(data_return_bytes, bytes))

        # decode
        returned_dict = MinosAvroProtocol.decode(data_return_bytes)
        self.assertEqual("get", returned_dict["headers"]["action"])
        self.assertEqual(123, returned_dict["headers"]["id"])
        self.assertEqual("testing", returned_dict["body"][0])
        self.assertEqual("testing2", returned_dict["body"][1])

    def test_encoder_decoder_without_body(self):
        headers = {"id": 123, "action": "get"}
        data_return_bytes = MinosAvroProtocol.encode(headers)
        self.assertEqual(True, isinstance(data_return_bytes, bytes))

        # decode
        returned_dict = MinosAvroProtocol.decode(data_return_bytes)
        self.assertEqual("get", returned_dict["headers"]["action"])
        self.assertEqual(123, returned_dict["headers"]["id"])

    def test_encode_raises(self):
        with self.assertRaises(MinosProtocolException):
            MinosAvroProtocol.encode(3)

    def test_decode_raises(self):
        with self.assertRaises(MinosProtocolException):
            MinosAvroValuesDatabase.decode(bytes())


class TestMinosAvroValuesDatabase(unittest.TestCase):
    def test_encoder_decoder_database_values_string(self):
        value = "String Test"
        data_return_bytes = MinosAvroValuesDatabase.encode(value)
        self.assertEqual(True, isinstance(data_return_bytes, bytes))

        # decode
        returned_string = MinosAvroValuesDatabase.decode(data_return_bytes)
        self.assertEqual("String Test", returned_string)

    def test_encoder_decoder_database_values_dictionary(self):
        value = {"first": "this is a string", "second": 123}
        data_return_bytes = MinosAvroValuesDatabase.encode(value)
        self.assertEqual(True, isinstance(data_return_bytes, bytes))

        # decode
        returned_string = MinosAvroValuesDatabase.decode(data_return_bytes)
        self.assertEqual("this is a string", returned_string["first"])
        self.assertEqual(123, returned_string["second"])

    def test_encoder_decoder_database_values_list(self):
        value = ["first", "second"]
        data_return_bytes = MinosAvroValuesDatabase.encode(value)
        self.assertEqual(True, isinstance(data_return_bytes, bytes))

        # decode
        returned_string = MinosAvroValuesDatabase.decode(data_return_bytes)
        self.assertEqual("first", returned_string[0])
        self.assertEqual("second", returned_string[1])

    def test_encode_raises(self):
        with self.assertRaises(MinosProtocolException):
            MinosAvroValuesDatabase.encode({}, schema={"": ""})

    def test_decode_raises(self):
        with self.assertRaises(MinosProtocolException):
            MinosAvroValuesDatabase.decode(bytes())


if __name__ == "__main__":
    unittest.main()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from .config import (
    BROKER,
    COMMAND,
    COMMANDS,
    DATABASE,
    ENDPOINT,
    EVENT,
    EVENTS,
    REST,
    SERVICE,
    MinosConfig,
    MinosConfigAbstract,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from pathlib import (
    Path,
)

BASE_PATH = Path(__file__).parent



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.

Unit test package for minos_microservice_common.
"""



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import abc
import typing as t

from ..exceptions import (
    MinosMessageException,
)
from ..protocol import (
    MinosAvroProtocol,
    MinosBinaryProtocol,
)


class MinosBaseResponse(abc.ABC):
    __slots__ = "binary_class", "_encoder", "_decoder", "_headers", "_body"

    @property
    def id(self):
        return self._headers["id"]

    @property
    def body(self):
        return self._body

    @property
    def headers(self):
        return self._headers

    @property
    def binary(self):
        if "id" not in self._headers:
            raise AttributeError("ID is mandatory, please use addId method before request binary data")

        if self._body:
            return self._encoder(self._headers, self._body)
        return self._encoder(self._headers)

    def addBody(self, body_val: t.Any):
        if isinstance(body_val, dict):
            self._body = {}
        elif isinstance(body_val, list):
            self._body = []
        self._body = body_val
        return self

    def addId(self, id: int):
        self._headers["id"] = id
        return self

    def sub_build(self):
        ...

    def build(self):
        """
        build the instance with all the default informations
        """
        self._prepare_binary_encoder_decoder()
        self._prepare_default_attributes()
        self.sub_build()

    def load(self, data: bytes):
        """
        import the bytes code and obtain the Request
        """
        self._prepare_binary_encoder_decoder()
        decoded_dict = self._decoder(data)
        self._headers = {}
        if "id" not in decoded_dict["headers"]:
            raise AttributeError("The Response dosn't have any ID configured, not valid!")

        self._headers = decoded_dict["headers"]
        if "body" in decoded_dict:
            self._body = decoded_dict["body"]

    def _prepare_binary_encoder_decoder(self):
        """
        configure the Encoder and the Decoder methods from the Binary Class Instance passed from the Builder
        """
        self._encoder = self.binary_class.encode
        self._decoder = self.binary_class.decode

    def _prepare_default_attributes(self):
        """
        instanciate the default class attributes like headers dict, body as None and the headers id
        """
        self._headers = {}
        self._body = None


class MinosRPCResponse(MinosBaseResponse):
    ...


class MinosResponse(object):
    @staticmethod
    def build(response_class: MinosBaseResponse, binary: MinosBinaryProtocol = MinosAvroProtocol):
        request_instance = response_class()
        if isinstance(request_instance, MinosBaseResponse):
            request_instance.binary_class = binary
            request_instance.build()
            return request_instance

        else:
            raise MinosMessageException("The request class must extend MinosBaseResponse")

    @staticmethod
    def load(data: bytes, response_class: MinosBaseResponse, binary: MinosBinaryProtocol = MinosAvroProtocol):
        """
        load binary data and convert in the Message Request Instance given
        """
        request_instance = response_class()
        if isinstance(request_instance, MinosBaseResponse):
            request_instance.binary_class = binary
            request_instance.load(data)
        return request_instance



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from typing import (
    Optional,
)

from minos.common import (
    Aggregate,
    ModelRef,
)


class Owner(Aggregate):
    """Aggregate ``Owner`` class for testing purposes."""

    name: str
    surname: str
    age: Optional[int]


class Car(Aggregate):
    """Aggregate ``Car`` class for testing purposes."""

    doors: int
    color: str
    # owner: Optional[ModelRef[Owner]]  # FIXME: There is a bug on the schema generation.



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import abc
import random
import typing as t

from ..exceptions import (
    MinosMessageException,
)
from ..logs import (
    log,
)
from ..protocol import (
    MinosAvroProtocol,
    MinosBinaryProtocol,
)


class MinosBaseRequest(abc.ABC):

    __slots__ = "binary_class", "_encoder", "_decoder", "_headers", "_body"

    @property
    def id(self):
        return self._headers["id"]

    @property
    def body(self):
        return self._body

    @property
    def headers(self):
        return self._headers

    @property
    def binary(self):
        if self._body:
            log.debug("imported with body")
            return self._encoder(self._headers, self._body)
        return self._encoder(self._headers)

    def addHeader(self, key: str, value: t.Any):
        self._headers[key] = value
        return self

    def addBody(self, body_val: t.Any):
        if isinstance(body_val, dict):
            self._body = {}
        elif isinstance(body_val, list):
            self._body = []
        self._body = body_val
        return self

    def addMethod(self, method: str):
        self._headers["method"] = method
        return self

    def addPath(self, path: str):
        self._headers["path"] = path
        return self

    def sub_build(self):
        ...

    def _prepare_binary_encoder_decoder(self):
        """
        configure the Encoder and the Decoder methods from the Binary Class Instance passed from the Builder
        """
        self._encoder = self.binary_class.encode
        self._decoder = self.binary_class.decode

    def _prepare_default_attributes(self):
        """
        instanciate the default class attributes like headers dict, body as None and the headers id
        """
        self._headers = {}
        self._body = None
        self._headers["id"] = random.randint(0, 1000)

    def build(self):
        """
        build the instance with all the default informations
        """
        self._prepare_binary_encoder_decoder()
        self._prepare_default_attributes()
        self.sub_build()

    def load(self, data: bytes):
        """
        import the bytes code and obtain the Request
        """
        self._prepare_binary_encoder_decoder()
        decoded_dict = self._decoder(data)
        self._headers = {}
        self._headers = decoded_dict["headers"]
        if "body" in decoded_dict:
            self._body = decoded_dict["body"]


class MinosRPCHeadersRequest(MinosBaseRequest):
    """
    Minos RPC Headers Request Class

    In the Minos RPC protocol the Headers are sent in a separete call.
    The Headers get the ID and the Action
    in the future more parameters are take in count
    """

    @property
    def action(self):
        return self._headers["action"]

    def addMethod(self, method: str):
        """
        the RPC does not support the classic method call
        """
        raise AttributeError("this method not exist")

    def addPath(self, path: str):
        """
        the RPC does not support the path
        """
        raise AttributeError("this method not exist")

    def addAction(self, action: str):
        """
        RPC support the action header parameter
        the action correspond to the RPC command
        """
        self._headers["action"] = action
        return self

    def sub_build(self):
        """
        instance more variables for this specific Message class
        """
        self._headers["type"] = "headers"


class MinosRPCBodyRequest(MinosBaseRequest):
    """
    Minos RPC Body Class

    The body are sent after the Headers.
    The Body have as Headers an ID, that correspond of the ID of the Headers call.
    and a set of atttributes that must be passed to the command function.
    """

    def addId(self, id: int):
        self._headers["id"] = id
        return self

    def addMethod(self, method: str):
        raise NotImplementedError

    def addPath(self, path: str):
        raise NotImplementedError

    def sub_build(self):
        """
        instance more variables for this specific Message class
        """
        self._headers["type"] = "body"


class MinosRequest(object):
    @staticmethod
    def build(request_clas: t.Type[MinosBaseRequest], binary: MinosBinaryProtocol = MinosAvroProtocol):
        """
        this is the Builder that create specific Requests Objects
        """
        request_instance = request_clas()
        if isinstance(request_instance, MinosBaseRequest):
            request_instance.binary_class = binary
            request_instance.build()
            return request_instance

        else:
            raise MinosMessageException("The request class must extend MinosBaseRequest")

    @staticmethod
    def load(data: bytes, request_class: t.Type[MinosBaseRequest], binary: MinosBinaryProtocol = MinosAvroProtocol):
        """
        load binary data and convert in the Message Request Instance given
        """
        request_instance = request_class()
        if isinstance(request_instance, MinosBaseRequest):
            request_instance.binary_class = binary
            request_instance.load(data)
        return request_instance



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import abc
import collections
import os
import typing as t
from pathlib import (
    Path,
)

import yaml

from ..exceptions import (
    MinosConfigDefaultAlreadySetException,
    MinosConfigException,
)

BROKER = collections.namedtuple("Broker", "host port")
DATABASE = collections.namedtuple("Database", "path name")
QUEUE = collections.namedtuple("Queue", "database user password host port records retry")
ENDPOINT = collections.namedtuple("Endpoint", "name route method controller action")
EVENT = collections.namedtuple("Event", "name controller action")
COMMAND = collections.namedtuple("Command", "name controller action")
SERVICE = collections.namedtuple("Service", "name")

EVENTS = collections.namedtuple("Events", "broker database items queue")
COMMANDS = collections.namedtuple("Commands", "broker database items queue")
REST = collections.namedtuple("Rest", "broker endpoints")

REPOSITORY = collections.namedtuple("Repository", "database user password host port")

_default: t.Optional[MinosConfigAbstract] = None


class MinosConfigAbstract(abc.ABC):
    __slots__ = "_services", "_path"

    def __init__(self, path: t.Union[Path, str]):
        if isinstance(path, Path):
            path = str(path)
        self._services = {}
        self._path = path
        self._load(path)

    @abc.abstractmethod
    def _load(self, path: str) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def _get(self, key: str, **kwargs: t.Any):
        raise NotImplementedError

    def _file_exit(self, path: str) -> bool:
        if os.path.isfile(path):
            return True
        return False

    def __enter__(self) -> MinosConfigAbstract:
        self.set_default(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> t.NoReturn:
        self.unset_default()

    @staticmethod
    def set_default(value: MinosConfigAbstract) -> t.NoReturn:
        """Set default config.

        :param value: Default config.
        :return: This method does not return anything.
        """
        if MinosConfigAbstract.get_default() is not None:
            raise MinosConfigDefaultAlreadySetException("There is already another config set as default.")
        global _default
        _default = value

    @classmethod
    def get_default(cls) -> MinosConfigAbstract:
        """Get default config.

        :return: A ``MinosConfigAbstract`` instance.
        """
        global _default
        return _default

    @staticmethod
    def unset_default() -> t.NoReturn:
        """Unset the default config.

        :return: This method does not return anything.
        """
        global _default
        _default = None


class MinosConfig(MinosConfigAbstract):
    __slots__ = ("_data",)

    def _load(self, path):
        if self._file_exit(path):
            with open(path) as f:
                self._data = yaml.load(f, Loader=yaml.FullLoader)
        else:
            raise MinosConfigException(f"Check if this path: {path} is correct")

    def _get(self, key: str, **kwargs: t.Any) -> t.Union[str, int, t.Dict[str, t.Any], None]:
        if key in self._data:
            return self._data[key]
        return None

    @property
    def service(self):
        service = self._get("service")
        return SERVICE(name=service["name"])

    @property
    def rest(self):
        rest_info = self._get("rest")
        broker = BROKER(host=rest_info["host"], port=rest_info["port"])
        endpoints = []
        for endpoint in rest_info["endpoints"]:
            endpoints.append(
                ENDPOINT(
                    name=endpoint["name"],
                    route=endpoint["route"],
                    method=endpoint["method"].upper(),
                    controller=endpoint["controller"],
                    action=endpoint["action"],
                )
            )
        return REST(broker=broker, endpoints=endpoints)

    @property
    def events(self):
        event_info = self._get("events")
        broker = BROKER(host=event_info["broker"], port=event_info["port"])
        database = DATABASE(path=event_info["database"]["path"], name=event_info["database"]["name"])
        queue = QUEUE(
            database=event_info["queue"]["database"],
            user=event_info["queue"]["user"],
            password=event_info["queue"]["password"],
            host=event_info["queue"]["host"],
            port=event_info["queue"]["port"],
            records=event_info["queue"]["records"],
            retry=event_info["queue"]["retry"],
        )
        events = []
        for event in event_info["items"]:
            events.append(EVENT(name=event["name"], controller=event["controller"], action=event["action"]))
        return EVENTS(broker=broker, items=events, database=database, queue=queue)

    @property
    def commands(self):
        command_info = self._get("commands")
        broker = BROKER(host=command_info["broker"], port=command_info["port"])
        database = DATABASE(path=command_info["database"]["path"], name=command_info["database"]["name"])
        queue = QUEUE(
            database=command_info["queue"]["database"],
            user=command_info["queue"]["user"],
            password=command_info["queue"]["password"],
            host=command_info["queue"]["host"],
            port=command_info["queue"]["port"],
            records=command_info["queue"]["records"],
            retry=command_info["queue"]["retry"],
        )
        commands = []
        for command in command_info["items"]:
            commands.append(COMMAND(name=command["name"], controller=command["controller"], action=command["action"]))
        return COMMANDS(broker=broker, items=commands, database=database, queue=queue)

    @property
    def repository(self) -> t.NamedTuple:
        """Get the repository config.

        :return: A ``Repository`` NamedTuple instance.
        """
        command_info = self._get("repository")
        return REPOSITORY(
            database=os.getenv("POSTGRES_DATABASE", command_info["database"]),
            user=os.getenv("POSTGRES_USER", command_info["user"]),
            password=os.getenv("POSTGRES_PASSWORD", command_info["password"]),
            host=os.getenv("POSTGRES_HOST", command_info["host"]),
            port=os.getenv("POSTGRES_PORT", command_info["port"]),
        )



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from .request import (
    MinosBaseRequest,
    MinosRequest,
    MinosRPCBodyRequest,
    MinosRPCHeadersRequest,
)
from .response import (
    MinosBaseResponse,
    MinosResponse,
    MinosRPCResponse,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from typing import (
    Any,
    Optional,
)

import aiopg

from .abc import (
    MinosRepository,
)
from .entries import (
    MinosRepositoryEntry,
)


class PostgreSqlMinosRepository(MinosRepository):
    """PostgreSQL-based implementation of the repository class in ``minos``."""

    def __init__(
        self,
        host: str = None,
        port: int = None,
        database: str = None,
        user: str = None,
        password: str = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.host = host
        self.port = port
        self.database = database
        self.user = user
        self.password = password

    async def _setup(self):
        """Setup miscellaneous repository thing.

        In the PostgreSQL case, configures the needed table to be used to store the data.

        :return: This method does not return anything.
        """
        await self._create_events_table()

    async def _create_events_table(self):
        await self._submit_sql(_CREATE_ACTION_ENUM_QUERY, fetch=False)
        await self._submit_sql(_CREATE_TABLE_QUERY, fetch=False)

    async def _submit(self, entry: MinosRepositoryEntry) -> MinosRepositoryEntry:
        params = {
            "action": entry.action.value,
            "aggregate_id": entry.aggregate_id,
            "aggregate_name": entry.aggregate_name,
            "data": entry.data,
        }
        response = await self._submit_sql(_INSERT_VALUES_QUERY, params)
        entry.id = response[0][0]
        entry.aggregate_id = response[0][1]
        entry.version = response[0][2]
        return entry

    async def _select(
        self, aggregate_id: int = None, aggregate_name: str = None, *args, **kwargs,
    ) -> list[MinosRepositoryEntry]:
        if aggregate_id is None and aggregate_name is None:
            response = await self._submit_sql(_SELECT_ALL_ENTRIES_QUERY)
            entries = [MinosRepositoryEntry(*row) for row in response]
        else:
            params = (aggregate_id, aggregate_name)
            response = await self._submit_sql(_SELECT_ENTRIES_QUERY, params)
            entries = [MinosRepositoryEntry(aggregate_id, aggregate_name, *row) for row in response]

        return entries

    async def _submit_sql(self, query: str, *args, fetch: bool = True, **kwargs) -> Optional[list[tuple[Any, ...]]]:
        async with self._connection() as connect:
            async with connect.cursor() as cursor:
                await cursor.execute(query, *args, **kwargs)
                if not fetch:
                    return None
                return await cursor.fetchall()

    def _connection(self):
        return aiopg.connect(
            host=self.host, port=self.port, dbname=self.database, user=self.user, password=self.password,
        )


_CREATE_ACTION_ENUM_QUERY = """
DO
$$
    BEGIN
        IF NOT EXISTS(SELECT *
                      FROM pg_type typ
                               INNER JOIN pg_namespace nsp
                                          ON nsp.oid = typ.typnamespace
                      WHERE nsp.nspname = current_schema()
                        AND typ.typname = 'action_type') THEN
            CREATE TYPE action_type AS ENUM ('insert', 'update', 'delete');
        END IF;
    END;
$$
LANGUAGE plpgsql;
""".strip()

_CREATE_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS events (
    id BIGSERIAL PRIMARY KEY,
    action ACTION_TYPE NOT NULL,
    aggregate_id BIGINT NOT NULL,
    aggregate_name TEXT NOT NULL,
    version INT NOT NULL,
    data BYTEA NOT NULL,
    UNIQUE (aggregate_id, aggregate_name, version)
);
""".strip()

_INSERT_VALUES_QUERY = """
INSERT INTO events (id, action, aggregate_id, aggregate_name, version, data)
VALUES (default,
        %(action)s,
        (
            CASE %(aggregate_id)s
                WHEN 0 THEN (
                    SELECT (CASE COUNT(*) WHEN 0 THEN 1 ELSE MAX(aggregate_id) + 1 END)
                    FROM events
                    WHERE aggregate_name = %(aggregate_name)s
                )
                ELSE %(aggregate_id)s END
            ),
        %(aggregate_name)s,
        (
            SELECT (CASE COUNT(*) WHEN 0 THEN 1 ELSE MAX(version) + 1 END)
            FROM events
            WHERE aggregate_id = %(aggregate_id)s
              AND aggregate_name = %(aggregate_name)s
        ),
        %(data)s)
RETURNING id, aggregate_id, version;
""".strip()

_SELECT_ENTRIES_QUERY = """
SELECT version, data, id, action
FROM events
WHERE aggregate_id = %s AND aggregate_name = %s;
""".strip()

_SELECT_ALL_ENTRIES_QUERY = """
SELECT aggregate_id, aggregate_name, version, data, id, action
FROM events
""".strip()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import abc


class MinosStorage(abc.ABC):
    @abc.abstractmethod
    def add(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def build(self, **kwargs):
        raise NotImplementedError



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from typing import (
    Optional,
)

from minos.common import (
    MinosModel,
    MissingSentinel,
    ModelRef,
)


class Base(MinosModel):
    """
    base class derived directly from MinosModel
    """

    id: int

    @staticmethod
    def validate_id(value: int) -> bool:
        """Validate non negative ids."""
        return value >= 0


class User(Base):
    """
    Class for Inheritance Test
    """

    username: Optional[str]

    @staticmethod
    def parse_username(value: Optional[str]) -> Optional[str]:
        """Parse username into a cleaner format.

        :param value: username to be parsed.
        :return: An string object.
        """
        if value is None or value is MissingSentinel:
            return None
        return value.lower()

    def validate_username(self, value: str) -> bool:
        """Validate that the username is ``None`` or a single word."""
        if self.id == 0 and value != "admin":
            return False
        return not value.count(" ")


class ShoppingList(MinosModel):
    """Class to test ``MinosModel`` composition."""

    user: Optional[ModelRef[User]]
    cost: float

    def parse_cost(self, value: Optional[str]) -> float:
        """Parse a number encoded as string with a semicolon as decimal separator.

        :param value: cost to be parsed.
        :return: A float value.
        """
        if self.user is not None:
            if self.user.username == "admin":
                return 0.0

        if value is None or value is MissingSentinel:
            return float("inf")
        if isinstance(value, float):
            return value
        return float(value.replace(".", "").replace(",", "."))


class Analytics(Base):
    """Class to test ``MinosModel`` recursive type validation."""

    orders: dict[str, list[ModelRef[ShoppingList]]]


class Customer(User):
    """
    Test a Model Class with List
    """

    name: Optional[str]
    surname: Optional[str]
    is_admin: Optional[bool]
    lists: Optional[list[int]]

    @staticmethod
    def parse_name(name: str) -> Optional[str]:
        """Parse name into a cleaner format.

        :param name: name to be parsed.
        :return: An string object.
        """
        if name is None or name is MissingSentinel:
            return None
        return name.title()


class CustomerDict(User):
    """
    Test a Model Class with Dictionary
    """

    name: str
    surname: str
    friends: dict[str, str]


class CustomerFailList(User):
    """
    Test a Model Class with a List wrong formatted
    """

    name: Optional[str]
    surname: Optional[str]
    listes_failed: Optional[list]


class CustomerFailDict(User):
    """
    Test a Model Class with a Dictionary wrong formatted
    """

    name: Optional[str]
    surname: Optional[str]
    friends: dict



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import shutil

import pytest
from minos.common import (
    MinosStorageLmdb,
)


@pytest.fixture
def path():
    return "./tests/test_db.lmdb"


@pytest.fixture(autouse=True)
def clear_database(path):
    yield
    # Code that will run after your test, for example:
    shutil.rmtree(path, ignore_errors=True)


def test_storage_add_text(path):
    minosEnv = MinosStorageLmdb.build(path)
    minosEnv.add("TestOne", "first", "Text Value")

    value_returned = minosEnv.get("TestOne", "first")
    assert value_returned == "Text Value"


def test_storage_add_int(path):
    minosEnv = MinosStorageLmdb.build(path)
    minosEnv.add("TestOne", "first", 123)

    value_returned = minosEnv.get("TestOne", "first")
    assert value_returned == 123


def test_storage_add_dict(path):
    minosEnv = MinosStorageLmdb.build(path)
    minosEnv.add("TestOne", "first", {"key_one": "hello", "key_two": "minos"})

    value_returned = minosEnv.get("TestOne", "first")
    assert value_returned["key_one"] == "hello"
    assert value_returned["key_two"] == "minos"


def test_storage_add_multi_dict(path):
    minosEnv = MinosStorageLmdb.build(path)
    minosEnv.add("TestOne", "first", {"key_one": "hello", "key_two": {"sub_key": "this is a sub text"}})

    value_returned = minosEnv.get("TestOne", "first")
    assert value_returned["key_one"] == "hello"
    assert value_returned["key_two"]["sub_key"] == "this is a sub text"


def test_storage_add_list(path):
    minosEnv = MinosStorageLmdb.build(path)
    minosEnv.add("TestOne", "first", ["hello", "minos"])

    value_returned = minosEnv.get("TestOne", "first")
    assert value_returned[0] == "hello"
    assert value_returned[1] == "minos"


def test_storage_add_multi_table(path):
    minosEnv = MinosStorageLmdb.build(path)
    minosEnv.add("TestOne", "first", "Text Value")
    minosEnv.add("TestTwo", "first_double", "Text Double Value")
    minosEnv.add("TestTwo", "first", "Text Value Diff")

    value_returned = minosEnv.get("TestOne", "first")
    assert value_returned == "Text Value"

    value_returned = minosEnv.get("TestTwo", "first_double")
    assert value_returned == "Text Double Value"

    value_returned = minosEnv.get("TestTwo", "first")
    assert value_returned == "Text Value Diff"


def test_storage_delete(path):
    minosEnv = MinosStorageLmdb.build(path)
    minosEnv.add("TestOne", "first", "Text Value")
    minosEnv.add("TestOne", "second", "Text Second Value")

    minosEnv.delete("TestOne", "first")
    value_returned = minosEnv.get("TestOne", "second")
    assert value_returned == "Text Second Value"

    exception_value = minosEnv.get("TestOne", "first")
    assert exception_value is None


def test_storage_update(path):
    minosEnv = MinosStorageLmdb.build(path)
    minosEnv.add("TestOne", "first", "Text Value")

    value_returned = minosEnv.get("TestOne", "first")
    assert value_returned == "Text Value"

    minosEnv.update("TestOne", "first", "Updated Text Value")
    updated_value = minosEnv.get("TestOne", "first")
    assert updated_value == "Updated Text Value"



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
import unittest

from minos.common import (
    MinosRepositoryAction,
    MinosRepositoryEntry,
    MinosRepositoryUnknownActionException,
)


class TestMinosRepositoryAction(unittest.TestCase):
    def test_value_of(self):
        self.assertEqual(MinosRepositoryAction.INSERT, MinosRepositoryAction.value_of("insert"))
        self.assertEqual(MinosRepositoryAction.UPDATE, MinosRepositoryAction.value_of("update"))
        self.assertEqual(MinosRepositoryAction.DELETE, MinosRepositoryAction.value_of("delete"))

    def test_value_of_raises(self):
        with self.assertRaises(MinosRepositoryUnknownActionException):
            MinosRepositoryAction.value_of("foo")


class TestMinosRepositoryEntry(unittest.TestCase):
    def test_constructor(self):
        entry = MinosRepositoryEntry(1234, "example.Car", 0, bytes("car", "utf-8"))
        self.assertEqual(1234, entry.aggregate_id)
        self.assertEqual("example.Car", entry.aggregate_name)
        self.assertEqual(0, entry.version)
        self.assertEqual(bytes("car", "utf-8"), entry.data)
        self.assertEqual(None, entry.id)
        self.assertEqual(None, entry.action)

    def test_constructor_extended(self):
        entry = MinosRepositoryEntry(1234, "example.Car", 0, bytes("car", "utf-8"), 5678, MinosRepositoryAction.INSERT)
        self.assertEqual(1234, entry.aggregate_id)
        self.assertEqual("example.Car", entry.aggregate_name)
        self.assertEqual(0, entry.version)
        self.assertEqual(bytes("car", "utf-8"), entry.data)
        self.assertEqual(5678, entry.id)
        self.assertEqual(MinosRepositoryAction.INSERT, entry.action)

    def test_id_set(self):
        entry = MinosRepositoryEntry(1234, "example.Car", 0, bytes("car", "utf-8"))
        self.assertEqual(None, entry.id)
        entry.id = 5678
        self.assertEqual(5678, entry.id)

    def test_id_action(self):
        entry = MinosRepositoryEntry(1234, "example.Car", 0, bytes("car", "utf-8"))
        self.assertEqual(None, entry.action)
        entry.action = MinosRepositoryAction.INSERT
        self.assertEqual(MinosRepositoryAction.INSERT, entry.action)


if __name__ == "__main__":
    unittest.main()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
import unittest

from minos.common import (
    EmptyMinosModelSequenceException,
    MinosAttributeValidationException,
    MinosConfigDefaultAlreadySetException,
    MinosConfigException,
    MinosException,
    MinosMalformedAttributeException,
    MinosModelAttributeException,
    MinosModelException,
    MinosParseAttributeException,
    MinosRepositoryAggregateNotFoundException,
    MinosRepositoryDeletedAggregateException,
    MinosRepositoryException,
    MinosRepositoryManuallySetAggregateIdException,
    MinosRepositoryManuallySetAggregateVersionException,
    MinosRepositoryNonProvidedException,
    MinosRepositoryUnknownActionException,
    MinosReqAttributeException,
    MinosTypeAttributeException,
    MultiTypeMinosModelSequenceException,
)


class TestExceptions(unittest.TestCase):
    def test_type(self):
        self.assertTrue(issubclass(MinosException, Exception))

    def test_base_repr(self):
        exception = MinosException("test")
        self.assertEqual("MinosException(message='test')", repr(exception))

    def test_base_str(self):
        exception = MinosException("test")
        self.assertEqual("test", str(exception))

    def test_config(self):
        self.assertTrue(issubclass(MinosConfigException, MinosException))

    def test_config_default_already_set(self):
        self.assertTrue(issubclass(MinosConfigDefaultAlreadySetException, MinosConfigException))

    def test_repository_aggregate_not_found(self):
        self.assertTrue(issubclass(MinosRepositoryAggregateNotFoundException, MinosRepositoryException))

    def test_repository_deleted_aggregate(self):
        self.assertTrue(issubclass(MinosRepositoryDeletedAggregateException, MinosRepositoryException))

    def test_repository_manually_set_aggregate_id(self):
        self.assertTrue(issubclass(MinosRepositoryManuallySetAggregateIdException, MinosRepositoryException))

    def test_repository_manually_set_aggregate_version(self):
        self.assertTrue(issubclass(MinosRepositoryManuallySetAggregateVersionException, MinosRepositoryException))

    def test_repository_bad_action(self):
        self.assertTrue(issubclass(MinosRepositoryUnknownActionException, MinosRepositoryException))

    def test_repository_non_set(self):
        self.assertTrue(issubclass(MinosRepositoryNonProvidedException, MinosRepositoryException))

    def test_model(self):
        self.assertTrue(issubclass(MinosModelException, MinosException))

    def test_model_emtpy_sequence(self):
        self.assertTrue(issubclass(EmptyMinosModelSequenceException, MinosModelException))

    def test_model_multi_type_sequence(self):
        self.assertTrue(issubclass(MultiTypeMinosModelSequenceException, MinosModelException))

    def test_model_attribute(self):
        self.assertTrue(issubclass(MinosModelAttributeException, MinosException))

    def test_required_attribute(self):
        self.assertTrue(issubclass(MinosReqAttributeException, MinosModelAttributeException))

    def test_type_attribute(self):
        self.assertTrue(issubclass(MinosTypeAttributeException, MinosModelAttributeException))

    def test_type_attribute_repr(self):
        exception = MinosTypeAttributeException("foo", float, True)
        message = (
            "MinosTypeAttributeException(message=\"The <class 'float'> expected type for 'foo' "
            "does not match with the given data type: <class 'bool'>\")"
        )
        self.assertEqual(message, repr(exception))

    def test_malformed_attribute(self):
        self.assertTrue(issubclass(MinosMalformedAttributeException, MinosModelAttributeException))

    def test_parse_attribute(self):
        self.assertTrue(issubclass(MinosParseAttributeException, MinosModelAttributeException))

    def test_attribute_parse_repr(self):
        exception = MinosParseAttributeException("foo", 34, ValueError())
        message = (
            'MinosParseAttributeException(message="ValueError() '
            "was raised while parsing 'foo' field with 34 value.\")"
        )
        self.assertEqual(message, repr(exception))

    def test_attribute_validation(self):
        self.assertTrue(issubclass(MinosAttributeValidationException, MinosModelAttributeException))

    def test_attribute_validation_repr(self):
        exception = MinosAttributeValidationException("foo", 34)
        message = "MinosAttributeValidationException(message=\"34 value does not pass the 'foo' field validation.\")"
        self.assertEqual(message, repr(exception))


if __name__ == "__main__":
    unittest.main()



import pytest
from minos.common.broker import (
    MinosBaseBroker,
)


class MinosBroker(MinosBaseBroker):
    def __init__(self):
        self.database = self._database()

    def _database(self):
        pass

    def send(self):
        pass


def test_minos_base_broker():
    assert MinosBroker().send() is None



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
import unittest

from minos.common import (
    MinosInMemoryRepository,
    MinosRepository,
    MinosRepositoryAction,
    MinosRepositoryEntry,
)


class TestStringMethods(unittest.IsolatedAsyncioTestCase):
    def test_constructor(self):
        repository = MinosInMemoryRepository()
        self.assertIsInstance(repository, MinosRepository)

    async def test_insert(self):
        async with MinosInMemoryRepository() as repository:
            await repository.insert(MinosRepositoryEntry(0, "example.Car", 1, bytes("foo", "utf-8")))

            expected = [
                MinosRepositoryEntry(1, "example.Car", 1, bytes("foo", "utf-8"), 1, MinosRepositoryAction.INSERT)
            ]
            self.assertEqual(expected, await repository.select())

    async def test_update(self):
        async with MinosInMemoryRepository() as repository:
            await repository.update(MinosRepositoryEntry(0, "example.Car", 1, bytes("foo", "utf-8")))
            expected = [
                MinosRepositoryEntry(0, "example.Car", 1, bytes("foo", "utf-8"), 1, MinosRepositoryAction.UPDATE)
            ]
            self.assertEqual(expected, await repository.select())

    async def test_delete(self):
        async with MinosInMemoryRepository() as repository:
            repository = MinosInMemoryRepository()
            await repository.delete(MinosRepositoryEntry(0, "example.Car", 1, bytes()))
            expected = [MinosRepositoryEntry(0, "example.Car", 1, bytes(), 1, MinosRepositoryAction.DELETE)]
            self.assertEqual(expected, await repository.select())

    async def test_select(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 1, bytes("foo", "utf-8"), 1, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(1, "example.Car", 2, bytes("bar", "utf-8"), 2, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(2, "example.Car", 1, bytes("hello", "utf-8"), 3, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(1, "example.Car", 3, bytes("foobar", "utf-8"), 4, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.Car", 4, bytes(), 5, MinosRepositoryAction.DELETE),
            MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8"), 6, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.MotorCycle", 1, bytes("one", "utf-8"), 7, MinosRepositoryAction.INSERT),
        ]
        self.assertEqual(expected, await repository.select())

    async def test_select_empty(self):
        repository = MinosInMemoryRepository()
        self.assertEqual([], await repository.select())

    async def test_select_id(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 2, bytes("bar", "utf-8"), 2, MinosRepositoryAction.UPDATE),
        ]
        self.assertEqual(expected, await repository.select(id=2))

    async def test_select_id_lt(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 1, bytes("foo", "utf-8"), 1, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(1, "example.Car", 2, bytes("bar", "utf-8"), 2, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(2, "example.Car", 1, bytes("hello", "utf-8"), 3, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(1, "example.Car", 3, bytes("foobar", "utf-8"), 4, MinosRepositoryAction.UPDATE),
        ]
        self.assertEqual(expected, await repository.select(id_lt=5))

    async def test_select_id_gt(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 4, bytes(), 5, MinosRepositoryAction.DELETE),
            MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8"), 6, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.MotorCycle", 1, bytes("one", "utf-8"), 7, MinosRepositoryAction.INSERT),
        ]
        self.assertEqual(expected, await repository.select(id_gt=4))

    async def test_select_id_le(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 1, bytes("foo", "utf-8"), 1, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(1, "example.Car", 2, bytes("bar", "utf-8"), 2, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(2, "example.Car", 1, bytes("hello", "utf-8"), 3, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(1, "example.Car", 3, bytes("foobar", "utf-8"), 4, MinosRepositoryAction.UPDATE),
        ]
        self.assertEqual(expected, await repository.select(id_le=4))

    async def test_select_id_ge(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 4, bytes(), 5, MinosRepositoryAction.DELETE),
            MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8"), 6, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.MotorCycle", 1, bytes("one", "utf-8"), 7, MinosRepositoryAction.INSERT),
        ]
        self.assertEqual(expected, await repository.select(id_ge=5))

    async def test_select_aggregate_id(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(2, "example.Car", 1, bytes("hello", "utf-8"), 3, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8"), 6, MinosRepositoryAction.UPDATE),
        ]
        self.assertEqual(expected, await repository.select(aggregate_id=2))

    async def test_select_aggregate_name(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.MotorCycle", 1, bytes("one", "utf-8"), 7, MinosRepositoryAction.INSERT),
        ]
        self.assertEqual(expected, await repository.select(aggregate_name="example.MotorCycle"))

    async def test_select_version(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 4, bytes(), 5, MinosRepositoryAction.DELETE),
        ]
        self.assertEqual(expected, await repository.select(version=4))

    async def test_select_version_lt(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 1, bytes("foo", "utf-8"), 1, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(2, "example.Car", 1, bytes("hello", "utf-8"), 3, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(1, "example.MotorCycle", 1, bytes("one", "utf-8"), 7, MinosRepositoryAction.INSERT),
        ]
        self.assertEqual(expected, await repository.select(version_lt=2))

    async def test_select_version_gt(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 2, bytes("bar", "utf-8"), 2, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.Car", 3, bytes("foobar", "utf-8"), 4, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.Car", 4, bytes(), 5, MinosRepositoryAction.DELETE),
            MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8"), 6, MinosRepositoryAction.UPDATE),
        ]
        self.assertEqual(expected, await repository.select(version_gt=1))

    async def test_select_version_le(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 1, bytes("foo", "utf-8"), 1, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(2, "example.Car", 1, bytes("hello", "utf-8"), 3, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(1, "example.MotorCycle", 1, bytes("one", "utf-8"), 7, MinosRepositoryAction.INSERT),
        ]
        self.assertEqual(expected, await repository.select(version_le=1))

    async def test_select_version_ge(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 2, bytes("bar", "utf-8"), 2, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.Car", 3, bytes("foobar", "utf-8"), 4, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.Car", 4, bytes(), 5, MinosRepositoryAction.DELETE),
            MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8"), 6, MinosRepositoryAction.UPDATE),
        ]
        self.assertEqual(expected, await repository.select(version_ge=2))

    async def test_select_combine(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 3, bytes("foobar", "utf-8"), 4, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.Car", 4, bytes(), 5, MinosRepositoryAction.DELETE),
            MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8"), 6, MinosRepositoryAction.UPDATE),
        ]
        self.assertEqual(expected, await repository.select(aggregate_name="example.Car", id_ge=4))

    @staticmethod
    async def _build_repository():
        repository = MinosInMemoryRepository()
        await repository.insert(MinosRepositoryEntry(1, "example.Car", 1, bytes("foo", "utf-8")))
        await repository.update(MinosRepositoryEntry(1, "example.Car", 2, bytes("bar", "utf-8")))
        await repository.insert(MinosRepositoryEntry(2, "example.Car", 1, bytes("hello", "utf-8")))
        await repository.update(MinosRepositoryEntry(1, "example.Car", 3, bytes("foobar", "utf-8")))
        await repository.delete(MinosRepositoryEntry(1, "example.Car", 4))
        await repository.update(MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8")))
        await repository.insert(MinosRepositoryEntry(1, "example.MotorCycle", 1, bytes("one", "utf-8")))

        return repository


if __name__ == "__main__":
    unittest.main()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import abc
import typing as t


class MinosBinaryProtocol(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def encode(cls, headers: t.Dict, body: t.Any) -> bytes:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def decode(cls, data: bytes):
        raise NotImplementedError



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from .abstract import (
    MinosBinaryProtocol,
)
from .avro import (
    MinosAvroProtocol,
    MinosAvroValuesDatabase,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import logging

from aiomisc.log import (
    basic_config,
)

basic_config(level=logging.DEBUG, buffered=False, log_format="color")
log = logging.getLogger(__name__)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import unittest

import aiopg
from minos.common import (
    MinosInMemoryRepository,
    MinosRepository,
    MinosRepositoryAction,
    MinosRepositoryEntry,
    PostgreSqlMinosRepository,
)
from tests.aggregate_classes import (
    Car,
)
from tests.database_testcase import (
    PostgresAsyncTestCase,
)


class TestPostgreSqlMinosRepository(PostgresAsyncTestCase):
    def test_constructor(self):
        repository = PostgreSqlMinosRepository("host", 1234, "database", "user", "password")
        self.assertIsInstance(repository, MinosRepository)
        self.assertEqual("host", repository.host)
        self.assertEqual(1234, repository.port)
        self.assertEqual("database", repository.database)
        self.assertEqual("user", repository.user)
        self.assertEqual("password", repository.password)

    async def test_setup(self):
        async with aiopg.connect(**self.kwargs) as connection:
            async with connection.cursor() as cursor:
                template = "SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = 'events');"
                await cursor.execute(template.format(**self.kwargs))
                response = (await cursor.fetchone())[0]
                self.assertFalse(response)

        repository = PostgreSqlMinosRepository(**self.kwargs)
        await repository._setup()

        async with aiopg.connect(**self.kwargs) as connection:
            async with connection.cursor() as cursor:
                template = "SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = 'events');"
                await cursor.execute(template.format(**self.kwargs))
                response = (await cursor.fetchone())[0]
                self.assertTrue(response)

    async def test_aggregate(self):
        async with PostgreSqlMinosRepository(**self.kwargs) as repository:
            car = await Car.create(doors=3, color="blue", _repository=repository)
            await car.update(color="red")
            await car.update(doors=5)

            another = await Car.get_one(car.id, _repository=repository)
            self.assertEqual(car, another)

            await car.delete()

    async def test_insert(self):
        async with PostgreSqlMinosRepository(**self.kwargs) as repository:
            await repository.insert(MinosRepositoryEntry(0, "example.Car", 1, bytes("foo", "utf-8")))

            expected = [
                MinosRepositoryEntry(1, "example.Car", 1, bytes("foo", "utf-8"), 1, MinosRepositoryAction.INSERT)
            ]
            self.assertEqual(expected, await repository.select())

    async def test_update(self):
        async with PostgreSqlMinosRepository(**self.kwargs) as repository:
            await repository.update(MinosRepositoryEntry(0, "example.Car", 1, bytes("foo", "utf-8")))
            expected = [
                MinosRepositoryEntry(0, "example.Car", 1, bytes("foo", "utf-8"), 1, MinosRepositoryAction.UPDATE)
            ]
            self.assertEqual(expected, await repository.select())

    async def test_delete(self):
        async with PostgreSqlMinosRepository(**self.kwargs) as repository:
            await repository.delete(MinosRepositoryEntry(0, "example.Car", 1, bytes()))
            expected = [MinosRepositoryEntry(0, "example.Car", 1, bytes(), 1, MinosRepositoryAction.DELETE)]
            self.assertEqual(expected, await repository.select())

    async def test_select(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(1, "example.Car", 1, bytes("foo", "utf-8"), 1, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(1, "example.Car", 2, bytes("bar", "utf-8"), 2, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(2, "example.Car", 1, bytes("hello", "utf-8"), 3, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(1, "example.Car", 3, bytes("foobar", "utf-8"), 4, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.Car", 4, bytes(), 5, MinosRepositoryAction.DELETE),
            MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8"), 6, MinosRepositoryAction.UPDATE),
            MinosRepositoryEntry(1, "example.MotorCycle", 1, bytes("one", "utf-8"), 7, MinosRepositoryAction.INSERT),
        ]
        self.assertEqual(expected, await repository.select())

    async def test_select_empty(self):
        repository = PostgreSqlMinosRepository(**self.kwargs)
        self.assertEqual([], await repository.select())

    async def test_select_filtered(self):
        repository = await self._build_repository()
        expected = [
            MinosRepositoryEntry(2, "example.Car", 1, bytes("hello", "utf-8"), 3, MinosRepositoryAction.INSERT),
            MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8"), 6, MinosRepositoryAction.UPDATE),
        ]
        self.assertEqual(expected, await repository.select(aggregate_name="example.Car", aggregate_id=2))

    async def _build_repository(self):
        repository = PostgreSqlMinosRepository(**self.kwargs)
        await repository._setup()
        await repository.insert(MinosRepositoryEntry(1, "example.Car", 1, bytes("foo", "utf-8")))
        await repository.update(MinosRepositoryEntry(1, "example.Car", 2, bytes("bar", "utf-8")))
        await repository.insert(MinosRepositoryEntry(2, "example.Car", 1, bytes("hello", "utf-8")))
        await repository.update(MinosRepositoryEntry(1, "example.Car", 3, bytes("foobar", "utf-8")))
        await repository.delete(MinosRepositoryEntry(1, "example.Car", 4))
        await repository.update(MinosRepositoryEntry(2, "example.Car", 2, bytes("bye", "utf-8")))
        await repository.insert(MinosRepositoryEntry(1, "example.MotorCycle", 1, bytes("one", "utf-8")))

        return repository


if __name__ == "__main__":
    unittest.main()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import typing as t

import lmdb

from ..protocol import (
    MinosAvroValuesDatabase,
)
from .abstract import (
    MinosStorage,
)


def _encode_values(value: t.Any) -> bytes:
    return MinosAvroValuesDatabase.encode(value)


def _decode_values(value: t.Any) -> bytes:
    return MinosAvroValuesDatabase.decode(value)


class MinosStorageLmdb(MinosStorage):
    __slots__ = "_env", "_tables"

    def __init__(self, env: lmdb.Environment):
        self._env: lmdb.Environment = env
        self._tables = {}

    @property
    def env(self) -> lmdb.Environment:
        return self._env

    def add(self, table: str, key: str, value: t.Any) -> t.NoReturn:
        db_instance = self._get_table(table)
        with self._env.begin(write=True) as txn:
            value_bytes: bytes = _encode_values(value)
            txn.put(key.encode(), value_bytes, db=db_instance)

    def get(self, table: str, key) -> t.Union[None, t.Any]:
        db_instance = self._get_table(table)
        with self._env.begin(db=db_instance) as txn:
            value_binary = txn.get(key.encode())
            if value_binary is not None:
                # decode the returned value
                return _decode_values(value_binary)
            return None

    def delete(self, table: str, key: str) -> t.NoReturn:
        db_instance = self._get_table(table)
        with self._env.begin(write=True, db=db_instance) as txn:
            txn.delete(key.encode())

    def update(self, table: str, key: str, value: t.Any) -> t.NoReturn:
        db_instance = self._get_table(table)
        with self._env.begin(write=True, db=db_instance) as txn:
            value_bytes: bytes = _encode_values(value)
            txn.put(key.encode(), value_bytes, db=db_instance, overwrite=True)

    def _get_table(self, table: str) -> lmdb._Database:
        if table in self._tables:
            return self._tables[table]
        else:
            # create a new table
            self._tables[table] = self._env.open_db(table.encode())
            return self._tables[table]

    @staticmethod
    def build(path_db: str, max_db: int = 10) -> "MinosStorageLmdb":
        """
        prepare the database initialization
        """

        env: lmdb.Environment = lmdb.open(path_db, max_dbs=max_db)
        return MinosStorageLmdb(env)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import unittest

from minos.common import (
    MinosImportException,
    classname,
    import_module,
)


class TestImportlib(unittest.TestCase):
    def test_import_module(self):
        object_class = import_module("tests.ImportedModule.ImportedClassTest")
        self.assertEqual("tests.ImportedModule.ImportedClassTest", classname(object_class))

    def test_import_module_exception(self):
        with self.assertRaises(MinosImportException):
            import_module("tests.ImportedModuleFail.ImportedClassTest")

    def test_classname(self):
        self.assertEqual("builtins.int", classname(int))
        self.assertEqual("minos.common.exceptions.MinosImportException", classname(MinosImportException))


if __name__ == "__main__":
    unittest.main()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import io
import typing as t

from fastavro import (
    parse_schema,
    reader,
    writer,
)

from ..exceptions import (
    MinosProtocolException,
)
from ..logs import (
    log,
)
from .abstract import (
    MinosBinaryProtocol,
)


class MinosAvroProtocol(MinosBinaryProtocol):
    @classmethod
    def encode(cls, headers: t.Dict, body: t.Any = None) -> bytes:
        """
        encoder in avro
        all the headers are converted in fields with doble underscore name
        the body is a set fields coming from the data type.
        """

        # prepare the headers
        schema = {
            "type": "record",
            "namespace": "org.minos.protocol",
            "name": "message",
            "fields": [
                {
                    "name": "headers",
                    "type": {"type": "map", "values": ["string", "int", "bytes", "long", "float", "boolean"]},
                },
                {
                    "name": "body",
                    "type": [
                        "null",
                        "string",
                        "int",
                        "bytes",
                        "boolean",
                        "float",
                        "long",
                        {"type": "array", "items": ["int", "string", "bytes", "long", "boolean", "float"]},
                        {
                            "type": "map",
                            "values": [
                                "null",
                                "int",
                                "string",
                                "bytes",
                                "boolean",
                                "float",
                                "long",
                                {
                                    "type": "array",
                                    "items": [
                                        "string",
                                        "int",
                                        "bytes",
                                        "long",
                                        "boolean",
                                        "float",
                                        {
                                            "type": "map",
                                            "values": ["string", "int", "bytes", "long", "boolean", "float"],
                                        },
                                    ],
                                },
                                {
                                    "type": "map",
                                    "values": [
                                        "string",
                                        "int",
                                        "bytes",
                                        "long",
                                        "boolean",
                                        "float",
                                        {
                                            "type": "map",
                                            "values": ["string", "int", "bytes", "long", "boolean", "float"],
                                        },
                                    ],
                                },
                            ],
                        },
                    ],
                },
            ],
        }
        final_data = {}
        final_data["headers"] = headers
        if body:
            final_data["body"] = body

        if not isinstance(final_data, list):
            final_data = [final_data]

        return cls._encode(schema, final_data)

    @classmethod
    def _encode(cls, schema_val: dict[str, t.Any], final_data: list[dict[str, t.Any]]) -> bytes:
        try:
            schema_bytes = parse_schema(schema_val)
            with io.BytesIO() as file:
                writer(file, schema_bytes, final_data)
                file.seek(0)
                content_bytes = file.getvalue()
                return content_bytes
        except Exception:
            raise MinosProtocolException(
                "Error encoding data, check if is a correct Avro Schema and Datum is provided."
            )

    @classmethod
    def decode(cls, data: bytes) -> t.Dict:
        data_return: t.Dict = {}
        try:
            data_return["headers"] = {}
            for schema_dict in cls._decode(data):
                log.debug(f"Avro: get the request/response in dict format")
                data_return["headers"] = schema_dict["headers"]
                # check wich type is body
                if "body" in schema_dict:
                    if isinstance(schema_dict["body"], dict):
                        data_return["body"] = {}
                    elif isinstance(schema_dict["body"], list):
                        data_return["body"] = []
                    data_return["body"] = schema_dict["body"]
            return data_return
        except Exception:
            raise MinosProtocolException("Error decoding string, check if is a correct Avro Binary data")

    @classmethod
    def _decode(cls, data: bytes) -> list[dict[str, t.Any]]:
        with io.BytesIO(data) as file:
            return list(reader(file))


class MinosAvroValuesDatabase(MinosAvroProtocol):
    """Encoder/Decoder class for values to be stored on the database with avro format."""

    @classmethod
    def encode(cls, value: t.Any, schema: t.Dict = None) -> bytes:
        """Encoder in avro for database Values
        all the headers are converted in fields with double underscore name
        the body is a set fields coming from the data type.

        :param value: The data to be stored.
        :param schema: The schema relative to the data.
        :return: A bytes object.
        """

        # prepare the headers
        if not schema:
            schema_val = {
                "type": "record",
                "namespace": "org.minos.protocol.database",
                "name": "value",
                "fields": [
                    {
                        "name": "content",
                        "type": [
                            "null",
                            "string",
                            "int",
                            "bytes",
                            "boolean",
                            "float",
                            "long",
                            {"type": "array", "items": ["int", "string", "bytes", "long", "boolean", "float"]},
                            {
                                "type": "map",
                                "values": [
                                    "null",
                                    "int",
                                    "string",
                                    "bytes",
                                    "boolean",
                                    "float",
                                    "long",
                                    {
                                        "type": "array",
                                        "items": [
                                            "string",
                                            "int",
                                            "bytes",
                                            "long",
                                            "boolean",
                                            "float",
                                            {
                                                "type": "map",
                                                "values": ["string", "int", "bytes", "long", "boolean", "float"],
                                            },
                                        ],
                                    },
                                    {
                                        "type": "map",
                                        "values": [
                                            "string",
                                            "int",
                                            "bytes",
                                            "long",
                                            "boolean",
                                            "float",
                                            {
                                                "type": "map",
                                                "values": ["string", "int", "bytes", "long", "boolean", "float"],
                                            },
                                        ],
                                    },
                                ],
                            },
                        ],
                    }
                ],
            }
            final_data = {}
            final_data["content"] = value
        else:
            schema_val = schema
            final_data = value

        if not isinstance(final_data, list):
            final_data = [final_data]

        return cls._encode(schema_val, final_data)

    @classmethod
    def decode(
        cls, data: bytes, content_root: bool = True, flatten: bool = True
    ) -> t.Union[dict[str, t.Any], list[dict[str, t.Any]]]:
        """Decode the given bytes of data into a single dictionary or a sequence of dictionaries.

        :param data: A bytes object.
        :param content_root: If ``True``
        :param flatten: If ``True`` tries to return the values as flat as possible.
        :return: A dictionary or a list of dictionaries.
        """
        try:
            ans = list()
            for schema_dict in cls._decode(data):
                log.debug(f"Avro Database: get the values data")
                if content_root:
                    schema_dict = schema_dict["content"]
                ans.append(schema_dict)
            if flatten and len(ans) == 1:
                return ans[0]
            return ans
        except Exception:
            raise MinosProtocolException("Error decoding string, check if is a correct Avro Binary data")



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import pytest
from minos.common import (
    MinosAvroProtocol,
    MinosRequest,
    MinosResponse,
    MinosRPCBodyRequest,
    MinosRPCHeadersRequest,
    MinosRPCResponse,
)


def test_request_avro_request_get_binary_headers():
    class_request: MinosRPCHeadersRequest = MinosRequest.build(MinosRPCHeadersRequest).addAction("with_args")
    binary_data = class_request.binary
    imported_request_class: MinosRPCHeadersRequest = MinosRequest.load(binary_data, MinosRPCHeadersRequest)

    assert imported_request_class.action == "with_args"


def test_request_avro_build_request_headers():
    class_request: MinosRPCHeadersRequest = MinosRequest.build(MinosRPCHeadersRequest).addAction("with_args")

    assert class_request.action == "with_args"


def test_request_avro_import_request_headers():
    headers = {"id": 123, "action": "get"}
    data_return_bytes = MinosAvroProtocol.encode(headers)

    class_request: MinosRPCHeadersRequest = MinosRequest.load(data_return_bytes, MinosRPCHeadersRequest)

    assert class_request.id == 123


def test_request_avro_build_request_body():
    class_request: MinosRPCBodyRequest = MinosRequest.build(MinosRPCBodyRequest).addId(123)
    assert class_request.id == 123


def test_request_avro_import_request_body():
    headers = {"id": 123, "action": "get"}
    body = {"test": "this is a test"}
    data_return_bytes = MinosAvroProtocol.encode(headers, body)

    class_request: MinosRPCBodyRequest = MinosRequest.load(data_return_bytes, MinosRPCBodyRequest)

    assert class_request.body["test"] == "this is a test"


def test_request_avro_attributes_error_request():
    try:
        class_request: MinosRPCHeadersRequest = MinosRequest.build(MinosRPCHeadersRequest).addMethod("GET")
        assert False == True
    except AttributeError as e:
        assert True == True


def test_request_avro_import_response():
    headers = {"id": 123}
    body = {"test": "this is a response test"}
    data_return_bytes = MinosAvroProtocol.encode(headers, body)

    class_response: MinosRPCResponse = MinosResponse.load(data_return_bytes, MinosRPCResponse)

    assert class_response.body["test"] == "this is a response test"


def test_request_avro_build_response():
    class_response: MinosRPCResponse = MinosResponse.build(MinosRPCResponse).addId(123).addBody("Test Response Body")

    assert class_response.body == "Test Response Body"
    assert class_response.id == 123


def test_request_avro_error_response_id():
    try:
        class_response: MinosRPCResponse = MinosResponse.build(MinosRPCResponse).addBody("Test Response Body")
        class_response.binary
        assert False == True
    except AttributeError:
        assert True == True


def test_request_avro_get_binary_response():
    try:
        class_response: MinosRPCResponse = MinosResponse.build(MinosRPCResponse).addId(123).addBody(
            "Test Response Body"
        )
        binary_format = class_response.binary
        imported_response_class: MinosRPCResponse = MinosResponse.load(binary_format, MinosRPCResponse)
        assert imported_response_class.id == 123
        assert imported_response_class.body == "Test Response Body"
    except AttributeError:
        assert False == True



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
import unittest

from minos.common import (
    MinosConfig,
    MinosInMemoryRepository,
    MinosRepositoryAggregateNotFoundException,
    MinosRepositoryDeletedAggregateException,
    MinosRepositoryManuallySetAggregateIdException,
    MinosRepositoryManuallySetAggregateVersionException,
    MinosRepositoryNonProvidedException,
    PostgreSqlMinosRepository,
)
from tests.aggregate_classes import (
    Car,
)
from tests.database_testcase import (
    PostgresAsyncTestCase,
)
from tests.utils import (
    BASE_PATH,
)


class TestAggregate(unittest.IsolatedAsyncioTestCase):
    async def test_create(self):
        async with MinosInMemoryRepository() as repository:
            car = await Car.create(doors=3, color="blue", _repository=repository)
            self.assertEqual(Car(1, 1, 3, "blue"), car)

    async def test_create_raises(self):
        async with MinosInMemoryRepository() as repository:
            with self.assertRaises(MinosRepositoryManuallySetAggregateIdException):
                await Car.create(id=1, doors=3, color="blue", _repository=repository)
            with self.assertRaises(MinosRepositoryManuallySetAggregateVersionException):
                await Car.create(version=1, doors=3, color="blue", _repository=repository)
        with self.assertRaises(MinosRepositoryNonProvidedException):
            await Car.create(doors=3, color="blue")

    async def test_classname(self):
        self.assertEqual("tests.aggregate_classes.Car", Car.classname)

    async def test_get(self):
        async with MinosInMemoryRepository() as repository:
            originals = [
                await Car.create(doors=3, color="blue", _repository=repository),
                await Car.create(doors=3, color="red", _repository=repository),
                await Car.create(doors=5, color="blue", _repository=repository),
            ]
            recovered = await Car.get([o.id for o in originals], _repository=repository)

            self.assertEqual(originals, recovered)

    async def test_get_one(self):
        async with MinosInMemoryRepository() as repository:
            original = await Car.create(doors=3, color="blue", _repository=repository)
            recovered = await Car.get_one(original.id, _repository=repository)

            self.assertEqual(original, recovered)

    async def test_get_one_raises(self):
        async with MinosInMemoryRepository() as repository:
            with self.assertRaises(MinosRepositoryAggregateNotFoundException):
                await Car.get_one(0, _repository=repository)
        with self.assertRaises(MinosRepositoryNonProvidedException):
            await Car.get_one(1)

    async def test_update(self):
        async with MinosInMemoryRepository() as repository:
            car = await Car.create(doors=3, color="blue", _repository=repository)

            await car.update(color="red")
            self.assertEqual(Car(1, 2, 3, "red"), car)
            self.assertEqual(car, await Car.get_one(car.id, _repository=repository))

            await car.update(doors=5)
            self.assertEqual(Car(1, 3, 5, "red"), car)
            self.assertEqual(car, await Car.get_one(car.id, _repository=repository))

    async def test_update_raises(self):
        async with MinosInMemoryRepository() as repository:
            with self.assertRaises(MinosRepositoryManuallySetAggregateVersionException):
                await Car.update(version=1, _repository=repository)
        with self.assertRaises(MinosRepositoryNonProvidedException):
            await Car(1, 1, 3, "blue").update(1, doors=1)

    async def test_update_cls(self):
        async with MinosInMemoryRepository() as repository:
            car = await Car.create(doors=3, color="blue", _repository=repository)

            await Car.update(car.id, color="red", _repository=repository)
            self.assertEqual(Car(1, 2, 3, "red"), await Car.get_one(car.id, _repository=repository))

            await Car.update(car.id, doors=5, _repository=repository)
            self.assertEqual(Car(1, 3, 5, "red"), await Car.get_one(car.id, _repository=repository))

    async def test_refresh(self):
        async with MinosInMemoryRepository() as repository:
            car = await Car.create(doors=3, color="blue", _repository=repository)
            await Car.update(car.id, color="red", _repository=repository)
            await Car.update(car.id, doors=5, _repository=repository)

            self.assertEqual(Car(1, 1, 3, "blue"), car)
            await car.refresh()
            self.assertEqual(Car(1, 3, 5, "red"), car)

    async def test_delete(self):
        async with MinosInMemoryRepository() as repository:
            car = await Car.create(doors=3, color="blue", _repository=repository)
            await car.delete()
            with self.assertRaises(MinosRepositoryDeletedAggregateException):
                await Car.get_one(car.id, _repository=repository)

    async def test_delete_raises(self):
        with self.assertRaises(MinosRepositoryNonProvidedException):
            await Car(1, 1, 3, "blue").delete(1)

    async def test_delete_cls(self):
        async with MinosInMemoryRepository() as repository:
            car = await Car.create(doors=3, color="blue", _repository=repository)
            await Car.delete(car.id, _repository=repository)
            with self.assertRaises(MinosRepositoryDeletedAggregateException):
                await Car.get_one(car.id, _repository=repository)


class TestAggregateWithConfig(PostgresAsyncTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.config_file_path = BASE_PATH / "test_config.yaml"

    async def test_update(self):
        with MinosConfig(path=self.config_file_path):
            await PostgreSqlMinosRepository.from_config().setup()

            car = await Car.create(doors=3, color="blue")

            await car.update(color="red")
            self.assertEqual(Car(1, 2, 3, "red"), car)
            self.assertEqual(car, await Car.get_one(car.id))

            await car.update(doors=5)
            self.assertEqual(Car(1, 3, 5, "red"), car)
            self.assertEqual(car, await Car.get_one(car.id))

            await car.delete()
            with self.assertRaises(MinosRepositoryDeletedAggregateException):
                await Car.get_one(car.id)

            car = await Car.create(doors=3, color="blue")
            await Car.update(car.id, color="red")
            self.assertEqual(Car(2, 2, 3, "red"), await Car.get_one(car.id))

            await Car.delete(car.id)
            with self.assertRaises(MinosRepositoryDeletedAggregateException):
                await Car.get_one(car.id)


if __name__ == "__main__":
    unittest.main()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""


class ImportedClassTest(object):
    def return_test_example(self):
        return "test passed"



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
import unittest
from typing import (
    Optional,
    Union,
)

from minos.common import (
    MinosAttributeValidationException,
    MinosMalformedAttributeException,
    MinosReqAttributeException,
    MinosTypeAttributeException,
    ModelField,
    ModelRef,
)
from tests.modelClasses import (
    User,
)


class TestModelField(unittest.TestCase):
    def test_name(self):
        field = ModelField("test", int, 3)
        self.assertEqual("test", field.name)

    def test_type(self):
        field = ModelField("test", int, 3)
        self.assertEqual(int, field.type)

    def test_value_int(self):
        field = ModelField("test", int, 3)
        self.assertEqual(3, field.value)

    def test_value_float(self):
        field = ModelField("test", float, 3.14)
        self.assertEqual(3.14, field.value)

    def test_value_bytes(self):
        field = ModelField("test", bytes, bytes("foo", "utf-8"))
        self.assertEqual(bytes("foo", "utf-8"), field.value)

    def test_value_float_raises(self):
        with self.assertRaises(MinosTypeAttributeException):
            ModelField("test", float, [3])
        with self.assertRaises(MinosTypeAttributeException):
            ModelField("test", float, "foo")

    def test_value_bytes_raises(self):
        with self.assertRaises(MinosTypeAttributeException):
            ModelField("test", bytes, 3)

    def test_value_list_int(self):
        field = ModelField("test", list[int], [1, 2, 3])
        self.assertEqual([1, 2, 3], field.value)

    def test_value_list_str(self):
        field = ModelField("test", list[str], ["foo", "bar", "foobar"])
        self.assertEqual(["foo", "bar", "foobar"], field.value)

    def test_value_list_model_ref(self):
        field = ModelField("test", list[ModelRef[User]], [User(123), User(456)])
        self.assertEqual([User(123), User(456)], field.value)

    def test_avro_schema_int(self):
        field = ModelField("test", int, 1)
        expected = {"name": "test", "type": "int"}
        self.assertEqual(expected, field.avro_schema)

    def test_avro_schema_bool(self):
        field = ModelField("test", bool, True)
        expected = {"name": "test", "type": "boolean"}
        self.assertEqual(expected, field.avro_schema)

    def test_avro_schema_float(self):
        field = ModelField("test", float, 3.4)
        expected = {"name": "test", "type": "float"}
        self.assertEqual(expected, field.avro_schema)

    def test_avro_schema_string(self):
        field = ModelField("test", str, "foo")
        expected = {"name": "test", "type": "string"}
        self.assertEqual(expected, field.avro_schema)

    def test_avro_schema_bytes(self):
        field = ModelField("test", bytes, bytes("foo", "utf-8"))
        expected = {"name": "test", "type": "bytes"}
        self.assertEqual(expected, field.avro_schema)

    def test_avro_schema_dict(self):
        field = ModelField("test", dict[str, int], {"foo": 1, "bar": 2})
        expected = {"name": "test", "type": {"default": {}, "type": "map", "values": "int"}}
        self.assertEqual(expected, field.avro_schema)

    def test_avro_schema_list_model_ref(self):
        field = ModelField("test", list[Optional[ModelRef[User]]], [User(123), User(456)])
        expected = {
            "name": "test",
            "type": {
                "default": [],
                "items": [
                    {
                        "fields": [{"name": "id", "type": "int"}, {"name": "username", "type": ["string", "null"]}],
                        "name": "User",
                        "namespace": "tests.modelClasses",
                        "type": "record",
                    },
                    "null",
                ],
                "type": "array",
            },
        }
        self.assertEqual(expected, field.avro_schema)

    def test_avro_data_list_model_ref(self):
        field = ModelField("test", list[Optional[ModelRef[User]]], [User(123), User(456)])
        expected = [{"id": 123, "username": None}, {"id": 456, "username": None}]
        self.assertEqual(expected, field.avro_data)

    def test_avro_data_dict(self):
        field = ModelField("test", dict[str, int], {"foo": 1, "bar": 2})
        self.assertEqual({"bar": 2, "foo": 1}, field.avro_data)

    def test_avro_data_bytes(self):
        field = ModelField("test", bytes, bytes("foo", "utf-8"))
        self.assertEqual(b"foo", field.avro_data)

    def test_value_list_optional(self):
        field = ModelField("test", list[Optional[int]], [1, None, 3, 4])
        self.assertEqual([1, None, 3, 4], field.value)

    def test_value_list_raises(self):
        with self.assertRaises(MinosTypeAttributeException):
            ModelField("test", list[int], 3)

    def test_value_dict(self):
        field = ModelField("test", dict[str, bool], {"foo": True, "bar": False})
        self.assertEqual({"foo": True, "bar": False}, field.value)

    def test_value_dict_raises(self):
        with self.assertRaises(MinosTypeAttributeException):
            ModelField("test", dict[str, int], 3)

    def test_dict_keys_raises(self):
        with self.assertRaises(MinosMalformedAttributeException):
            ModelField("test", dict[int, int], {1: 2, 3: 4})

    def test_value_model_ref(self):
        user = User(1234)
        field = ModelField("test", ModelRef[User], user)
        self.assertEqual(user, field.value)

    def test_value_model_ref_raises(self):
        with self.assertRaises(MinosTypeAttributeException):
            ModelField("test", ModelRef[User], "foo")

    def test_value_model_ref_optional(self):
        field = ModelField("test", Optional[ModelRef[User]], None)
        self.assertIsNone(field.value)

        user = User(1234)
        field.value = user
        self.assertEqual(user, field.value)

    def test_value_unsupported(self):
        with self.assertRaises(MinosTypeAttributeException):
            ModelField("test", set[int], {3,})

    def test_value_setter(self):
        field = ModelField("test", int, 3)
        field.value = 3
        self.assertEqual(3, field.value)

    def test_value_setter_raises(self):
        field = ModelField("test", int, 3)
        with self.assertRaises(MinosReqAttributeException):
            field.value = None

    def test_value_setter_list_raises_required(self):
        with self.assertRaises(MinosReqAttributeException):
            ModelField("test", list[int])

    def test_value_setter_union_raises_required(self):
        with self.assertRaises(MinosReqAttributeException):
            ModelField("test", Union[int, str])

    def test_value_setter_dict(self):
        field = ModelField("test", dict[str, bool], {})
        field.value = {"foo": True, "bar": False}
        self.assertEqual({"foo": True, "bar": False}, field.value)

    def test_value_setter_dict_raises(self):
        field = ModelField("test", dict[str, bool], {})
        with self.assertRaises(MinosTypeAttributeException):
            field.value = "foo"
        with self.assertRaises(MinosTypeAttributeException):
            field.value = {"foo": 1, "bar": 2}
        with self.assertRaises(MinosTypeAttributeException):
            field.value = {1: True, 2: False}

    def test_empty_value_raises(self):
        with self.assertRaises(MinosReqAttributeException):
            ModelField("id", int)

    def test_union_type(self):
        with self.assertRaises(MinosReqAttributeException):
            ModelField("test", Union[int, list[int]], None)

    def test_optional_type(self):
        field = ModelField("test", Optional[int])
        self.assertEqual(Optional[int], field.type)

    def test_empty_optional_value(self):
        field = ModelField("test", Optional[int])
        self.assertEqual(None, field.value)

    def test_empty_field_equality(self):
        self.assertEqual(ModelField("test", Optional[int], None), ModelField("test", Optional[int]))

    def test_value_setter_optional_int(self):
        field = ModelField("test", Optional[int], 3)
        self.assertEqual(3, field.value)
        field.value = None
        self.assertEqual(None, field.value)
        field.value = 4
        self.assertEqual(4, field.value)

    def test_value_setter_optional_list(self):
        field = ModelField("test", Optional[list[int]], [1, 2, 3])
        self.assertEqual([1, 2, 3], field.value)
        field.value = None
        self.assertEqual(None, field.value)
        field.value = [4]
        self.assertEqual([4], field.value)

    def test_parser(self):
        parser = lambda x: x.title()
        field = ModelField("test", str, "foo", parser)
        self.assertEqual(parser, field.parser)

    def test_parser_non_set(self):
        field = ModelField("test", str, "foo")
        self.assertEqual(None, field.parser)

    def test_parser_value(self):
        field = ModelField("test", str, "foo", lambda x: x.title())
        self.assertEqual("Foo", field.value)

    def test_parser_with_casting(self):
        field = ModelField("test", float, "1.234,56", lambda x: x.replace(".", "").replace(",", "."))
        self.assertEqual(1234.56, field.value)

    def test_parser_optional(self):
        field = ModelField("test", Optional[str], None, lambda x: x if x is None else x.title())
        self.assertEqual(None, field.value)

    def test_validator(self):
        validator = lambda x: not x.count(" ")
        field = ModelField("test", str, "foo", validator=validator)
        self.assertEqual(validator, field.validator)

    def test_validator_non_set(self):
        field = ModelField("test", str, "foo")
        self.assertEqual(None, field.validator)

    def test_validator_ok(self):
        field = ModelField("test", str, "foo", validator=lambda x: not x.count(" "))
        self.assertEqual("foo", field.value)

    def test_validator_raises(self):
        with self.assertRaises(MinosAttributeValidationException):
            ModelField("test", str, "foo bar", validator=lambda x: not x.count(" "))

    def test_validator_optional(self):
        field = ModelField("test", Optional[str], validator=lambda x: not x.count(" "))
        self.assertEqual(None, field.value)

    def test_equal(self):
        self.assertEqual(ModelField("id", Optional[int], 3), ModelField("id", Optional[int], 3))
        self.assertNotEqual(ModelField("id", Optional[int], 3), ModelField("id", Optional[int], None))
        self.assertNotEqual(ModelField("id", Optional[int], 3), ModelField("foo", Optional[int], 3))
        self.assertNotEqual(ModelField("id", Optional[int], 3), ModelField("id", int, 3))

    def test_iter(self):
        self.assertEqual(("id", Optional[int], 3, None, None), tuple(ModelField("id", Optional[int], 3)))

    def test_hash(self):
        self.assertEqual(hash(("id", Optional[int], 3, None, None)), hash(ModelField("id", Optional[int], 3)))

    def test_repr(self):
        field = ModelField("test", Optional[int], 1, parser=lambda x: x * 10, validator=lambda x: x > 0)
        expected = "ModelField(name='test', type=typing.Optional[int], value=10, parser=<lambda>, validator=<lambda>)"
        self.assertEqual(expected, repr(field))

    def test_repr_empty_parser(self):
        field = ModelField("test", Optional[int], 1)
        self.assertEqual(
            "ModelField(name='test', type=typing.Optional[int], value=1, parser=None, validator=None)", repr(field),
        )


if __name__ == "__main__":
    unittest.main()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
import abc
from ..common import MinosModel, ModelRef, Aggregate


class MinosBaseBroker(abc.ABC):
    """Base Broker class."""

    @abc.abstractmethod
    def _database(self):  # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    def send(self):  # pragma: no cover
        raise NotImplementedError


class Event(MinosModel):
    """Base Event class."""

    topic: str
    model: str
    items: list[ModelRef[Aggregate]]


class Command(MinosModel):
    """Base Command class."""

    topic: str
    model: str
    items: list[ModelRef[Aggregate]]
    reply_on: str



#!/usr/bin/env python

"""The setup script."""

from setuptools import (
    find_namespace_packages,
    setup,
)

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.md") as history_file:
    history = history_file.read()

requirements = ["aiomisc", "six", "fastavro", "orjson", "lmdb", "pyyaml", "aiopg"]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
]

setup(
    author="Andrea Mucci",
    author_email="andrea@clariteia.com",
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Python Package with common Classes and Utilities used in Minos Microservices",
    install_requires=requirements,
    long_description_content_type="text/markdown",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="minos_microservice_common",
    name="minos_microservice_common",
    packages=find_namespace_packages(include=["minos.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    version="0.0.3",
    zip_safe=False,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
import unittest

from minos.common import (
    MinosConfig,
    MinosConfigAbstract,
    MinosConfigDefaultAlreadySetException,
    MinosConfigException,
)
from tests.utils import (
    BASE_PATH,
)


class TestMinosConfig(unittest.TestCase):
    def setUp(self) -> None:
        self.config_file_path = BASE_PATH / "test_config.yaml"

    def test_config_ini_fail(self):
        with self.assertRaises(MinosConfigException):
            MinosConfig(path=BASE_PATH / "test_fail_config.yaml")

    def test_config_service(self):
        config = MinosConfig(path=self.config_file_path)
        service = config.service
        self.assertEqual("Order", service.name)

    def test_config_rest(self):
        config = MinosConfig(path=self.config_file_path)
        rest = config.rest

        broker = rest.broker
        self.assertEqual("localhost", broker.host)
        self.assertEqual(8900, broker.port)

        endpoints = rest.endpoints
        self.assertEqual("AddOrder", endpoints[0].name)

    def test_config_events(self):
        config = MinosConfig(path=self.config_file_path)
        events = config.events
        broker = events.broker
        self.assertEqual("localhost", broker.host)
        self.assertEqual(9092, broker.port)

    def test_config_events_database(self):
        config = MinosConfig(path=self.config_file_path)
        events = config.events
        database = events.database
        self.assertEqual("./tests/local_db.lmdb", database.path)
        self.assertEqual("database_events_test", database.name)

    def test_config_events_queue_database(self):
        config = MinosConfig(path=self.config_file_path)
        events = config.events
        queue = events.queue
        self.assertEqual("broker_db", queue.database)
        self.assertEqual("broker", queue.user)
        self.assertEqual("br0k3r", queue.password)
        self.assertEqual("localhost", queue.host)
        self.assertEqual(5432, queue.port)
        self.assertEqual(10, queue.records)
        self.assertEqual(2, queue.retry)

    def test_config_commands_database(self):
        config = MinosConfig(path=self.config_file_path)
        commands = config.commands
        database = commands.database
        self.assertEqual("./tests/local_db.lmdb", database.path)
        self.assertEqual("database_commands_test", database.name)

    def test_config_commands_queue_database(self):
        config = MinosConfig(path=self.config_file_path)
        commands = config.commands
        queue = commands.queue
        self.assertEqual("broker_db", queue.database)
        self.assertEqual("broker", queue.user)
        self.assertEqual("br0k3r", queue.password)
        self.assertEqual("localhost", queue.host)
        self.assertEqual(5432, queue.port)
        self.assertEqual(10, queue.records)
        self.assertEqual(2, queue.retry)

    def test_get_default_default(self):
        with MinosConfig(path=self.config_file_path) as config:
            self.assertEqual(config, MinosConfigAbstract.get_default())

    def test_multiple_default_config_raises(self):
        with self.assertRaises(MinosConfigDefaultAlreadySetException):
            with MinosConfig(path=self.config_file_path):
                with MinosConfig(path=self.config_file_path):
                    pass


if __name__ == "__main__":
    unittest.main()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
import os
import unittest

import aiopg


class PostgresAsyncTestCase(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self._meta_kwargs = {
            "host": os.getenv("POSTGRES_HOST", "localhost"),
            "port": os.getenv("POSTGRES_PORT", 5432),
            "database": os.getenv("POSTGRES_DATABASE", "postgres"),
            "user": os.getenv("POSTGRES_USER", "postgres"),
            "password": os.getenv("POSTGRES_PASSWORD", "postgres"),
        }

        self.kwargs = self._meta_kwargs | {
            "database": "test_db",
            "user": "test_user",
            "password": "test_password",
        }

    async def asyncSetUp(self):
        async with aiopg.connect(**self._meta_kwargs) as connection:
            async with connection.cursor() as cursor:
                template = "DROP DATABASE IF EXISTS {database};"
                await cursor.execute(template.format(**self.kwargs))

                template = "DROP ROLE IF EXISTS {user};"
                await cursor.execute(template.format(**self.kwargs))

                template = "CREATE ROLE {user} WITH SUPERUSER CREATEDB LOGIN ENCRYPTED PASSWORD {password!r};"
                await cursor.execute(template.format(**self.kwargs))

                template = "CREATE DATABASE {database} WITH OWNER = {user};"
                await cursor.execute(template.format(**self.kwargs))

    async def asyncTearDown(self):
        async with aiopg.connect(**self._meta_kwargs) as connection:
            async with connection.cursor() as cursor:
                template = "DROP DATABASE IF EXISTS {database}"
                await cursor.execute(template.format(**self.kwargs))



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

import importlib
import typing as t

import six

from .exceptions import (
    MinosImportException,
)


def import_module(module: str) -> t.Type:
    """Import the given module from a package"""
    try:
        if "." in module:
            parts = module.split(".")
            name = ".".join(parts[:-1])

        module_ref = importlib.import_module(name)
        kallable = getattr(module_ref, parts[-1])
        if not six.callable(kallable):
            raise TypeError("The module is not callable")
        return kallable
    except ImportError as e:
        raise MinosImportException("Error importing Package")


def classname(cls: t.Type) -> str:
    """Compute the given class full name.

    :param cls: Target class.
    :return: An string object.
    """
    return f"{cls.__module__}.{cls.__qualname__}"



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
import unittest
from typing import (
    Optional,
)

import pytest
from minos.common import (
    EmptyMinosModelSequenceException,
    MinosAttributeValidationException,
    MinosMalformedAttributeException,
    MinosModelException,
    MinosParseAttributeException,
    MinosReqAttributeException,
    MinosTypeAttributeException,
    ModelField,
    MultiTypeMinosModelSequenceException,
)
from tests.modelClasses import (
    Analytics,
    Customer,
    CustomerFailDict,
    CustomerFailList,
    ShoppingList,
    User,
)


class TestMinosModel(unittest.TestCase):
    def test_constructor_args(self):
        model = Customer(1234, "johndoe", "John", "Doe")
        self.assertEqual(1234, model.id)
        self.assertEqual("johndoe", model.username)
        self.assertEqual("John", model.name)
        self.assertEqual("Doe", model.surname)

    def test_constructor_multiple_values(self):
        with self.assertRaises(TypeError):
            Customer(1234, id=1234)
        with self.assertRaises(TypeError):
            Customer(None, id=1234)

    def test_constructor_kwargs(self):
        model = Customer(id=1234, username="johndoe", name="John", surname="Doe")
        self.assertEqual(1234, model.id)
        self.assertEqual("johndoe", model.username)
        self.assertEqual("Doe", model.surname)
        self.assertEqual("John", model.name)

    def test_aggregate_setter(self):
        model = Customer(1234)
        model.name = "John"
        model.surname = "Doe"
        self.assertEqual(1234, model.id)
        self.assertEqual("Doe", model.surname)
        self.assertEqual("John", model.name)

    def test_attribute_parser_same_type(self):
        model = Customer(1234, name="john")
        self.assertEqual("John", model.name)

    def test_attribute_parser_different_type(self):
        model = ShoppingList(User(1234), cost="1.234,56")
        self.assertEqual(1234.56, model.cost)

    def test_attribute_parser_attribute(self):
        model = ShoppingList(User(1234, "admin"), cost="1.234,56")
        self.assertEqual(0.0, model.cost)

    def test_aggregate_int_as_string_type_setter(self):
        model = Customer("1234")
        model.name = "John"
        self.assertEqual(1234, model.id)
        self.assertEqual("John", model.name)

    def test_aggregate_wrong_int_type_setter(self):
        with pytest.raises(MinosTypeAttributeException):
            Customer("1234S")

    def test_aggregate_string_type_setter(self):
        model = Customer(123)
        model.name = "John"
        self.assertEqual("John", model.name)

    def test_aggregate_wrong_string_type_setter_with_parser(self):
        model = Customer(123)
        with pytest.raises(MinosParseAttributeException):
            model.name = 456

    def test_aggregate_wrong_string_type_setter(self):
        model = Customer(123)
        with pytest.raises(MinosTypeAttributeException):
            model.surname = 456

    def test_aggregate_bool_type_setter(self):
        model = Customer(123)
        model.name = "John"
        model.is_admin = True
        self.assertTrue(model.is_admin)

    def test_aggregate_wrong_bool_type_setter(self):
        model = Customer(123)
        model.name = "John"
        with pytest.raises(MinosTypeAttributeException):
            model.is_admin = "True"

    def test_aggregate_empty_mandatory_field(self):
        with pytest.raises(MinosReqAttributeException):
            Customer()

    def test_model_is_freezed_class(self):
        model = Customer(123)
        with pytest.raises(MinosModelException):
            model.address = "str kennedy"

    def test_model_list_class_attribute(self):
        model = Customer(123)
        model.lists = [1, 5, 8, 6]

        self.assertEqual([1, 5, 8, 6], model.lists)

    def test_model_list_wrong_attribute_type(self):
        model = Customer(123)
        with pytest.raises(MinosTypeAttributeException):
            model.lists = [1, "hola", 8, 6]

    def test_model_ref(self):
        shopping_list = ShoppingList(cost=3.14)
        user = User(1234)
        shopping_list.user = user
        self.assertEqual(user, shopping_list.user)

    def test_avro_schema(self):
        expected = {
            "fields": [
                {
                    "name": "user",
                    "type": [
                        {
                            "fields": [{"name": "id", "type": "int"}, {"name": "username", "type": ["string", "null"]}],
                            "name": "User",
                            "namespace": "tests.modelClasses",
                            "type": "record",
                        },
                        "null",
                    ],
                },
                {"name": "cost", "type": "float"},
            ],
            "name": "ShoppingList",
            "namespace": "tests.modelClasses",
            "type": "record",
        }
        self.assertEqual(expected, ShoppingList.avro_schema)

    def test_avro_data(self):
        shopping_list = ShoppingList(User(1234))
        expected = {"cost": float("inf"), "user": {"id": 1234, "username": None}}
        self.assertEqual(expected, shopping_list.avro_data)

    def test_avro_bytes(self):
        shopping_list = ShoppingList(User(1234))
        self.assertIsInstance(shopping_list.avro_bytes, bytes)

    def test_avro_schema_simple(self):
        customer = Customer(1234)
        expected = {
            "fields": [
                {"name": "id", "type": "int"},
                {"name": "username", "type": ["string", "null"]},
                {"name": "name", "type": ["string", "null"]},
                {"name": "surname", "type": ["string", "null"]},
                {"name": "is_admin", "type": ["boolean", "null"]},
                {"name": "lists", "type": [{"default": [], "items": "int", "type": "array"}, "null"]},
            ],
            "name": "Customer",
            "namespace": "tests.modelClasses",
            "type": "record",
        }
        self.assertEqual(expected, customer.avro_schema)

    def test_avro_data_simple(self):
        customer = Customer(1234)
        expected = {
            "id": 1234,
            "is_admin": None,
            "lists": None,
            "name": None,
            "surname": None,
            "username": None,
        }
        self.assertEqual(expected, customer.avro_data)

    def test_avro_bytes_single(self):
        customer = Customer(1234)
        avro_bytes = customer.avro_bytes
        self.assertIsInstance(avro_bytes, bytes)
        decoded_customer = Customer.from_avro_bytes(avro_bytes)
        self.assertEqual(customer, decoded_customer)

    def test_avro_bytes_sequence(self):
        customers = [Customer(1234), Customer(5678)]
        avro_bytes = Customer.to_avro_bytes(customers)
        self.assertIsInstance(avro_bytes, bytes)
        decoded_customer = Customer.from_avro_bytes(avro_bytes)
        self.assertEqual(customers, decoded_customer)

    def test_avro_bytes_empty_sequence(self):
        with self.assertRaises(EmptyMinosModelSequenceException):
            Customer.to_avro_bytes([])

    def test_avro_bytes_multi_type_sequence(self):
        with self.assertRaises(MultiTypeMinosModelSequenceException):
            Customer.to_avro_bytes([User(1234), Customer(5678)])

    def test_model_ref_raises(self):
        shopping_list = ShoppingList(cost=3.14)
        with self.assertRaises(MinosTypeAttributeException):
            shopping_list.user = "foo"

    def test_recursive_type_composition(self):
        orders = {
            "foo": [ShoppingList(User(1)), ShoppingList(User(1))],
            "bar": [ShoppingList(User(2)), ShoppingList(User(2))],
        }

        analytics = Analytics(1, orders)
        self.assertEqual(orders, analytics.orders)

    def test_model_fail_list_class_attribute(self):
        with pytest.raises(MinosMalformedAttributeException):
            CustomerFailList(123)

    def test_model_fail_dict_class_attribute(self):
        with pytest.raises(MinosMalformedAttributeException):
            CustomerFailDict(123)

    def test_empty_required_value(self):
        with self.assertRaises(MinosReqAttributeException):
            User()

    def test_validate_required_raises(self):
        with self.assertRaises(MinosAttributeValidationException):
            User(-34)

    def test_validate_optional_raises(self):
        with self.assertRaises(MinosAttributeValidationException):
            User(1234, username="user name")

    def test_fields(self):
        user = User(123)
        fields = {
            "id": ModelField("id", int, 123, validator=user.validate_id),
            "username": ModelField(
                "username", Optional[str], parser=user.parse_username, validator=user.validate_username
            ),
        }
        self.assertEqual(fields, user.fields)

    def test_equal(self):
        a, b = User(123), User(123)
        self.assertEqual(a, b)

    def test_complex_equal(self):
        a, b = ShoppingList(User(1234), cost="1.234,56"), ShoppingList(User(1234), cost="1.234,56")
        self.assertEqual(a, b)

    def test_not_equal(self):
        a, b = User(123), User(456)
        self.assertNotEqual(a, b)

    def test_iter(self):
        user = User(123)
        expected = {
            "id": ModelField("id", int, 123, validator=user.validate_id),
            "username": ModelField(
                "username", Optional[str], parser=user.parse_username, validator=user.validate_username
            ),
        }
        self.assertEqual(expected, dict(user))

    def test_hash(self):
        user = User(123)

        expected = hash(
            (
                ("id", ModelField("id", int, 123, validator=user.validate_id)),
                (
                    "username",
                    ModelField("username", Optional[str], parser=user.parse_username, validator=user.validate_username),
                ),
            )
        )
        self.assertEqual(expected, hash(user))

    def test_repr(self):
        shopping_list = ShoppingList(User(1234), cost="1.234,56")
        expected = (
            "ShoppingList(fields=[ModelField(name='user', type=typing.Optional[minos.common.model.abc.types.ModelRef"
            "[tests.modelClasses.User]], value=User(fields=[ModelField(name='id', type=<class 'int'>, value=1234, "
            "parser=None, validator=validate_id), ModelField(name='username', type=typing.Optional[str], value=None, "
            "parser=parse_username, validator=validate_username)]), parser=None, validator=None), ModelField(name="
            "'cost', type=<class 'float'>, value=1234.56, parser=parse_cost, validator=None)])"
        )
        self.assertEqual(expected, repr(shopping_list))

    def test_classname_cls(self):
        self.assertEqual("tests.modelClasses.Customer", Customer.classname)

    def test_classname_instance(self):
        model = Customer(1234, "johndoe", "John", "Doe")
        self.assertEqual("tests.modelClasses.Customer", model.classname)


if __name__ == "__main__":
    unittest.main()
